import * as React from "react";
import { observer } from "mobx-react-lite";
import { observable, runInAction } from "mobx";
import { get } from 'lodash'
import { AppRender } from "@cloudbase/lowcode-render";
import { createWidgets, retryDataBinds, WidgetsContext, bindFuncObj, createComputed, resolveComponentProps} from '/src/utils/index'
import getStateFn from "./lowcode/state.js";
import computed from "./lowcode/computed.js";
import lifecycle from "./lowcode/lifecycle.js";
// import { app, $page } from '../../../../app/global-api'
<% if(isSandbox) { %>
import { ErrorBoundary } from 'react-error-boundary'
import { createErrorFallback } from '/src/utils/error'
<% } %>

<% handlersImports.forEach(handler => { %>
import handler$<%= handler.name %> from "./lowcode/handler/<%= handler.name %>.js";
<% }) %>

// Import Components outof sandbox
<% let componentsMap = {}; useComponents.forEach(compItem => {%>
<%if(!isSandbox && compItem.entries && compItem.entries.components){ %>
import { <%= compItem.name %> as <%= compItem.var %> } from "/src/libraries/<%= compItem.moduleName %>@<%= compItem.version %>/<%= compItem.entries.components %>";
<%} else if(!isSandbox && compItem.entries && compItem.entries.entry){ %><% if(!componentsMap[compItem.moduleName]){ componentsMap[compItem.moduleName] = true %>
import <%= compItem.moduleNameVar %> from "/src/libraries/<%= compItem.moduleName %>@<%= compItem.version %>/<%= compItem.entries.entry %>";
<% } %>const { <%= compItem.name %>: <%= compItem.var %> } = <%= compItem.moduleNameVar %>.components
<% } else if(!isSandbox) { %>
import <%= compItem.var %> from "/src/libraries/<%= compItem.moduleName %>@<%= compItem.version %>/components/<%= compItem.name %>";
<% } %>
<%}) %>

import * as constObj from '../../libCommonRes/const'
import * as toolsObj from '../../libCommonRes/tools'

import "./lowcode/style.css";

class CompositeCompWrapper extends React.Component {

  $WEAPPS_COMP = {}

  componentDidUpdate() {
    runInAction(() => {
      const { data } = this.props
      for(let prop in data) {
        // 更新 propsData
        if (typeof data[prop] !== 'function') {
          this.propsData[prop] = data[prop]
        }
      }
    })
  }


  constructor(props) {
    super(props);

    this.compConfig = <%= JSON.stringify(compConfig, null, 2) %>

    // Import Components in sandbox
    <% useComponents.forEach(compItem => {%>
    <% if(isSandbox){ %>
    const <%= compItem.var %> = window["@weapps-materials-main-<%= compItem.moduleName %>"].components["<%= compItem.name %>"];
    <% } %>
    <%}) %>

    this.virtualFields = Object.assign({}, props.pageVirtualFields || {}, {
    <% useComponents.forEach(compItem => {%>
      "<%= compItem.key %>": React.forwardRef((props,ref) => <<%= compItem.var %> {...resolveComponentProps(props,  <%= compItem.isPlainProps? 1:0 %>)} ref={ref}/>),
    <%}) %>
    });

    // 挂载给到 $comp.props.events 使用
    this.events = (<%= emitEvents %>).reduce((obj, trigger) => {
      obj[trigger] = (eventData, domEvent) => {
        this.props.emit(trigger, eventData)
        // this.props.emit(trigger, eventData, domEvent);
        // this.node._listeners && this.node._listeners.emit(trigger, eventData, domEvent)
      };
      return obj;
    }, {});
    this.handler = this.$WEAPPS_COMP.handler = {
      <% handlersImports.forEach(handler => { %>
      <%= handler.name %>: handler$<%= handler.name %>.bind(this),
      <% }) %>
    };
    this.componentSchema = <%= componentSchema %>;
    const widgetContext = <%= widgets %>
    const dataBinds = this._dataBinds = <%= dataBinds %>
    const defaultProps = <%= JSON.stringify(defaultProps, null, 2) %>
    this.propsData = observable(Object.assign({}, defaultProps, this.props.data || {}))
    this.$WEAPPS_COMP.lib = { const: constObj, tools: toolsObj }
    this.$WEAPPS_COMP.props = { ...this.props, events: this.events, data: this.propsData }
    this.state = this.$WEAPPS_COMP.state = observable(getStateFn.bind(this)())
    this.computed = this.$WEAPPS_COMP.computed = createComputed(computed, this)
    this.node = this.$WEAPPS_COMP.node = this.createWidgetNode(this) || {}
    this.widgets = this.$WEAPPS_COMP.widgets = createWidgets(widgetContext, dataBinds)
    // widgets 内的 dataBinds 可能需要关联 widgets，需要重新执行 dataBinds
    retryDataBinds()
    Object.keys(this.widgets || {}).forEach(widgetId => {
      // 将实例 ownerWidget 挂到内部组件上。内部组件就可以通过 $comp.node.ownerWidget 获取到所在的组件实例
      const node = this.widgets[widgetId]
      if (Array.isArray(node)) {
        node.forEach(item => {
          item.ownerWidget = this.node;
          item.getOwnerWidget = () => this.node;
        })
      }
      else {
        node.ownerWidget = this.node
        node.getOwnerWidget = () => this.node
      }
    })

    this.pageListenerInstances = [];
    this.createCompAPI(this)
  }

  // 创建自身节点
  createWidgetNode(compThis) {
    // 当为数组时，需要判断自己属于 widgets 的哪一项
    const {
      compositeParent,
      forIndexes,
      id
    } = compThis.props
    let widgetData = compositeParent.$WEAPPS_COMP
      ? compositeParent.$WEAPPS_COMP.widgets[id]
      : compositeParent.$page.widgets[id]
    if(Array.isArray(widgetData)) {
      widgetData = widgetData.length > 0 ? get(widgetData, forIndexes) : {}
    }
    widgetData = widgetData || {}
    Object.assign(widgetData, {
        getConfig: () => compThis.compConfig,
        getOwnerWidget: () => compThis.node.ownerWidget
      }
    )

    return widgetData
  }

  createCompAPI(compThis) {
    const { forwardRef } = compThis.props || {};
    compThis.$WEAPPS_COMP = {
      _dataBinds: compThis._dataBinds,
      compConfig: compThis.compConfig,
      widgets: compThis.widgets,
      node: compThis.node,
      handler: compThis.handler,
      lib: { const: constObj, tools: toolsObj },
      get props() {
        const { forwardRef, ...restProps } = compThis.props || {};
        return {...restProps, events: compThis.events, data: compThis.propsData }
      },
      get state() {
        return compThis.state
      },
      get computed() {
        return compThis.computed
      },
      get _instanceRef() {
        return forwardRef
      },
      get methods() {
        return forwardRef.current?.methods
      },
      set methods(value) {
        if(!forwardRef.current) {
          forwardRef.current = {}
        }
        forwardRef.current.methods = value
        return value
      }
    };
  }

  componentDidMount() {
    const onAttached = lifecycle.onAttached && lifecycle.onAttached.bind(this)
    Promise.resolve(onAttached()).then(()=>{
      lifecycle.onReady && lifecycle.onReady.bind(this)()
    })
  }

  componentWillUnmount() {
    // this.node._listeners && this.node._listeners.clear()
    lifecycle.onDetached && lifecycle.onDetached.bind(this)()
    this.$WEAPPS_COMP._instanceRef.current = undefined
  }

  render() {
    return (
      <WidgetsContext.Provider value={{ parent: this }}>
        <AppRender
          className={this.props.className}
          virtualFields={this.virtualFields}
          componentSchema={this.componentSchema}
          codeContext={this}
        />
      </WidgetsContext.Provider>
    );
  }
}

CompositeCompWrapper.contextType = WidgetsContext

export default observer((props, _ref) => (
  <% if(isSandbox) { %>
  <ErrorBoundary FallbackComponent={createErrorFallback('<%= name %>', '<%= id %>')}>
  <% } %>
    <CompositeCompWrapper {...props}  forwardRef={_ref}></CompositeCompWrapper>
  <% if(isSandbox) { %>
  </ErrorBoundary>
  <% } %>
), { forwardRef: true });

